/* vi: set ts=2 shiftwidth=2 expandtab:
 *
 * Copyright (C) 2003-2006  Simon Baldwin and Mark J. Tilford
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of version 2 of the GNU General Public License
 * as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */

/*
 * Module notes:
 *
 * o Is the whole interpolation and ALR passes thing right?  There's no
 *   documentation on it, and it's not intuitively implemented in Adrift.
 *
 * o Is dissecting HTML tags the right thing to do?
 */

#include <assert.h>
#include <string.h>
#include <stdio.h>
#include <ctype.h>

#include "scare.h"
#include "scprotos.h"


/* Assorted definitions and constants. */
static const sc_uint32 PRINTFILTER_MAGIC = 0xb4736417;
static const sc_uint32 BUFFER_GROW_INCREMENT = 32;
static const sc_char PERCENT = '%';
static const sc_char NUL = '\0';
static const sc_char LESSTHAN = '<';
static const sc_char GREATERTHAN = '>';

/* Trace flag, set before running. */
static sc_bool pf_trace = FALSE;

/* Enumerated variable types. */
enum
{ VAR_INTEGER = 'I',
  VAR_STRING = 'S'
};

/*
 * Table tying HTML-like tag strings to enumerated tag types.  Since it's
 * scanned sequentially by strncmp(), it's ordered so that longer strings
 * come before shorter ones.  The <br> tag is missing because this is
 * handled separately, as a simple put of '\n'.
 */
struct html_tags_s
{
  const sc_char *const tag_string;
  const sc_uint32 tag;
};

static const struct html_tags_s HTML_TAGS[] = {
  {"bgcolour", SC_TAG_BGCOLOR}, {"bgcolor", SC_TAG_BGCOLOR},
  {"waitkey", SC_TAG_WAITKEY},
  {"center", SC_TAG_CENTER}, {"/center", SC_TAG_ENDCENTER},
  {"centre", SC_TAG_CENTER}, {"/centre", SC_TAG_ENDCENTER},
  {"right", SC_TAG_RIGHT}, {"/right", SC_TAG_ENDRIGHT},
  {"font", SC_TAG_FONT}, {"/font", SC_TAG_ENDFONT},
  {"wait", SC_TAG_WAIT}, {"cls", SC_TAG_CLS},
  {"i", SC_TAG_ITALICS}, {"/i", SC_TAG_ENDITALICS},
  {"b", SC_TAG_BOLD}, {"/b", SC_TAG_ENDBOLD},
  {"u", SC_TAG_UNDERLINE}, {"/u", SC_TAG_ENDUNDERLINE},
  {"c", SC_TAG_SCOLOR}, {"/c", SC_TAG_ENDSCOLOR},
  {NULL, SC_TAG_UNKNOWN}
};

/*
 * Printfilter structure definition.  It defines a buffer for output,
 * associated size and length, and a note of any conversion to apply to
 * the next buffered character.
 */
typedef struct sc_printfilter_s
{
  sc_uint32 magic;
  sc_uint32 buffer_length;
  sc_uint32 buffer_size;
  sc_char *buffer;
  sc_bool new_sentence;
} sc_printfilter_t;


/*
 * pf_create()
 *
 * Create and return a new printfilter.
 */
sc_printfilterref_t
pf_create (void)
{
  sc_printfilterref_t filter;

  /* Create a new printfilter. */
  filter = sc_malloc (sizeof (*filter));
  filter->magic = PRINTFILTER_MAGIC;
  filter->buffer_length = 0;
  filter->buffer_size = 0;
  filter->buffer = NULL;
  filter->new_sentence = FALSE;

  /* Return the new printfilter. */
  return filter;
}


/*
 * pf_destroy()
 *
 * Destroy a printfilter and free its allocated memory.
 */
void
pf_destroy (sc_printfilterref_t filter)
{
  assert (filter && filter->magic == PRINTFILTER_MAGIC);

  /* Free any buffer space. */
  sc_free (filter->buffer);

  /* Shred and free the printfilter. */
  memset (filter, 0, sizeof (*filter));
  sc_free (filter);
}


/*
 * pf_interpolate()
 *
 * Replace %...% elements in a string by their variable values.  Return a
 * malloc'ed string with replacements done, and set changed if any variables
 * were found and replaced.
 *
 * If a %...% element exists that is not a variable, then it's left in as is.
 * Similarly, an unmatched (single) % in a string is also left as is.  There
 * appears to be no facility in the file format for escaping literal '%'
 * characters, and since some games have strings with this character in them,
 * this is probably all that can be done.
 */
static sc_char *
pf_interpolate (const sc_char *string,
                sc_var_setref_t vars, sc_bool *changed)
{
  sc_char *output_buffer, *variable_ptr, *variable_name;
  const sc_char *string_ptr;

  /* Initially, no changes made. */
  *changed = FALSE;

  /* Size variable_name to the same size as the input string. */
  variable_name = sc_malloc (strlen (string) + 1);

  /* Start with an empty output buffer. */
  output_buffer = sc_malloc (1);
  strcpy (output_buffer, "");

  /* Run through the string looking for variables. */
  string_ptr = string;
  for (variable_ptr = strchr (string_ptr, PERCENT);
       variable_ptr; variable_ptr = strchr (string_ptr, PERCENT))
    {
      sc_char type;
      sc_vartype_t vt_rvalue;

      /* Append buffer up to percent character. */
      output_buffer = sc_realloc (output_buffer,
                                  strlen (output_buffer)
                                  + variable_ptr - string_ptr + 1);
      strncat (output_buffer, string_ptr, variable_ptr - string_ptr);

      /* Get the variable name. */
      if (sscanf (variable_ptr, "%%%[^%]%%", variable_name) != 1)
        {
          /*
           * Mismatched % -- append the single % character and continue.
           */
          output_buffer = sc_realloc (output_buffer,
                                      strlen (output_buffer) + 2);
          strncat (output_buffer, variable_ptr, 1);
          string_ptr = variable_ptr + 1;
          continue;
        }

      /* Get the variable value, expected to exist. */
      if (!var_get (vars, variable_name, &type, &vt_rvalue))
        {
          /*
           * The %...% substring wasn't a known variable name -- as above,
           * append the single % character and continue.
           */
          output_buffer = sc_realloc (output_buffer,
                                      strlen (output_buffer) + 2);
          strncat (output_buffer, variable_ptr, 1);
          string_ptr = variable_ptr + 1;
          continue;
        }

      /* Get variable value and append to the string. */
      switch (type)
        {
        case VAR_INTEGER:
          {
            sc_char buffer[16];

            sprintf (buffer, "%ld", vt_rvalue.integer);
            output_buffer = sc_realloc (output_buffer,
                                        strlen (output_buffer)
                                        + strlen (buffer) + 1);
            strcat (output_buffer, buffer);
            break;
          }

        case VAR_STRING:
          output_buffer = sc_realloc (output_buffer,
                                      strlen (output_buffer)
                                      + strlen (vt_rvalue.string) + 1);
          strcat (output_buffer, vt_rvalue.string);
          break;

        default:
          sc_fatal ("pf_interpolate: invalid type, %d\n", type);
        }

      /* Advance string_ptr over the %...% variable name. */
      string_ptr = variable_ptr + strlen (variable_name) + 2;

      /* Note that a variable was found. */
      *changed = TRUE;
    }

  /* Append remainder of the string. */
  output_buffer = sc_realloc (output_buffer,
                              strlen (output_buffer) + strlen (string_ptr) + 1);
  strcat (output_buffer, string_ptr);

  /* Free temporary variable_name, and return constructed string. */
  sc_free (variable_name);
  return output_buffer;
}


/*
 * pf_replace_alrs()
 *
 * Replace any ALRs found in the string with their equivalents.  Return
 * a malloc'ed string with replacements done, and set changed if any
 * variables were found and replaced.
 */
static sc_char *
pf_replace_alrs (const sc_char *string, sc_prop_setref_t bundle,
                 sc_bool *changed, sc_bool alr_applied[])
{
  sc_vartype_t vt_key[3], vt_rvalue;
  sc_int32 alr_count, index_;
  sc_char *output_buffer, *temporary_buffer;

  /* Initially, no changes made. */
  *changed = FALSE;

  /* Get a count of ALRs. */
  vt_key[0].string = "ALRs";
  alr_count =
    (prop_get (bundle, "I<-s", &vt_rvalue, vt_key)) ? vt_rvalue.integer : 0;

  /*
   * Start with the output buffer being a copy of string, and the temporary
   * buffer being an empty string.
   */
  output_buffer = sc_malloc (strlen (string) + 1);
  strcpy (output_buffer, string);
  temporary_buffer = sc_malloc (1);
  strcpy (temporary_buffer, "");

  /* Run through each ALR that exists. */
  for (index_ = 0; index_ < alr_count; index_++)
    {
      sc_int32 alr;
      const sc_char *original, *replacement;
      sc_char *alr_ptr, *string_ptr, *swap_buffer;

      /* Ignore ALR indexes that have already been applied. */
      if (alr_applied[index_])
        continue;

      /*
       * Get the actual ALR number for the ALR.  This comes from the index
       * that we sorted earlier by length of original string.
       */
      vt_key[0].string = "ALRs2";
      vt_key[1].integer = index_;
      vt_key[2].string = "ALRIndex";
      alr = prop_get_integer (bundle, "I<-sis", vt_key);

      /* Set this ALR in the properties key. */
      vt_key[0].string = "ALRs";
      vt_key[1].integer = alr;

      /* Retrieve the ALR original string, and NULL replacement. */
      vt_key[2].string = "Original";
      original = prop_get_string (bundle, "S<-sis", vt_key);
      replacement = NULL;

      /* Ignore pathological empty originals. */
      if (original[0] == NUL)
        continue;

      /* Switch output and temporary buffers. */
      swap_buffer = temporary_buffer;
      temporary_buffer = output_buffer;
      output_buffer = swap_buffer;

      /* Empty the output buffer string. */
      output_buffer = sc_realloc (output_buffer, 1);
      strcpy (output_buffer, "");

      /* Run through the temporary looking for things to replace. */
      string_ptr = temporary_buffer;
      for (alr_ptr = strstr (string_ptr, original);
           alr_ptr; alr_ptr = strstr (string_ptr, original))
        {
          /* Append buffer up to original string start. */
          output_buffer = sc_realloc (output_buffer,
                                      strlen (output_buffer)
                                      + alr_ptr - string_ptr + 1);
          strncat (output_buffer, string_ptr, alr_ptr - string_ptr);

          /*
           * Optimize by retrieving the replacement string only on demand --
           * NULL means not yet retrieved.
           */
          if (!replacement)
            {
              vt_key[2].string = "Replacement";
              replacement = prop_get_string (bundle, "S<-sis", vt_key);
            }

          /* Append the replacement string. */
          output_buffer = sc_realloc (output_buffer,
                                      strlen (output_buffer)
                                      + strlen (replacement) + 1);
          strcat (output_buffer, replacement);

          /* Advance string_ptr over the original ALR string found. */
          string_ptr = alr_ptr + strlen (original);

          /* Note that a replacement was made. */
          alr_applied[index_] = TRUE;
          *changed = TRUE;
        }

      /* See if string_ptr advanced, that is, was ALR found? */
      if (string_ptr != temporary_buffer)
        {
          /* Append remainder of the string. */
          output_buffer = sc_realloc (output_buffer,
                                      strlen (output_buffer)
                                      + strlen (string_ptr) + 1);
          strcat (output_buffer, string_ptr);
        }
      else
        {
          /*
           * ALR not found, so optimize by just switching the buffers around
           * again.
           */
          swap_buffer = temporary_buffer;
          temporary_buffer = output_buffer;
          output_buffer = swap_buffer;
        }
    }

  /* Free temporary buffer memory. */
  sc_free (temporary_buffer);

  /* Return constructed string. */
  return output_buffer;
}


/*
 * pf_output_text()
 *
 * Edit the tag-stripped text element passed in, substituting &lt; &gt;
 * with < >, then send to the OS-specific output functions.
 */
static void
pf_output_text (const sc_char *string)
{
  sc_uint32 index_;
  sc_char *copy;

  /* Do nothing if the string element is empty. */
  if (strlen (string) == 0)
    return;

  /* Copy the string. */
  copy = sc_malloc (strlen (string) + 1);
  strcpy (copy, string);

  /* Run along the copy replacing &..; elements by equivalents. */
  for (index_ = 0; copy[index_] != NUL; index_++)
    {
      if (sc_strncasecmp (copy + index_, "&lt;", strlen ("&lt;")) == 0)
        {
          copy[index_] = LESSTHAN;
          memmove (copy + index_ + 1,
                   copy + index_ + 4, strlen (copy) - index_ - 3);
        }
      else if (sc_strncasecmp (copy + index_, "&gt;", strlen ("&gt;")) == 0)
        {
          copy[index_] = GREATERTHAN;
          memmove (copy + index_ + 1,
                   copy + index_ + 4, strlen (copy) - index_ - 3);
        }
    }

  /* Print out, then free, the modified string copy. */
  if_print_string (copy);
  sc_free (copy);
}


/*
 * pf_output_tag()
 *
 * Output an HTML-like tag element to the OS-specific tag handling function.
 */
static void
pf_output_tag (const sc_char *tag)
{
  sc_uint16 index_;
  sc_uint32 tag_code;
  const sc_char *tag_argument;

  /* For a simple <br> tag, just print out a newline. */
  if (sc_strcasecmp (tag, "br") == 0 && (tag[2] == NUL || isspace (tag[2])))
    {
      if_print_character ('\n');
      return;
    }

  /* Look up the tag in the tags table. */
  tag_code = SC_TAG_UNKNOWN;
  tag_argument = tag;
  for (index_ = 0; HTML_TAGS[index_].tag_string; index_++)
    {
      if (sc_strncasecmp (tag, HTML_TAGS[index_].tag_string,
                          strlen (HTML_TAGS[index_].tag_string)) == 0)
        {
          const sc_char *next;

          /*
           * Found a possible table match -- check that it's a full match, and
           * if it is, set tag_code and tag_argument and break.
           */
          next = tag + strlen (HTML_TAGS[index_].tag_string);
          if (*next == NUL || isspace (*next))
            {
              while (*next != NUL && isspace (*next))
                next++;
              tag_code = HTML_TAGS[index_].tag;
              tag_argument = next;
              break;
            }
        }
    }

  /*
   * Call tag output.  Tag_arg is "" if none, the tag argument if any, or the
   * the complete tag if no match.
   */
  if_print_tag (tag_code, tag_argument);
}


/*
 * pf_output_untagged()
 *
 * Break apart HTML-like string into normal text elements, and HTML-like
 * tags.
 */
static void
pf_output_untagged (const sc_char *string)
{
  sc_char *temporary_buffer, *untagged_text, *tag, *tag_ptr;
  const sc_char *string_ptr;

  /* Create a general temporary string. */
  temporary_buffer = sc_malloc (strlen (string) + 1);
  untagged_text = tag = temporary_buffer;

  /* Run through the string looking for <...> tags. */
  string_ptr = string;
  for (tag_ptr = strchr (string_ptr, LESSTHAN);
       tag_ptr; tag_ptr = strchr (string_ptr, LESSTHAN))
    {
      /* Handle characters up to the tag start. */
      strncpy (untagged_text, string_ptr, tag_ptr - string_ptr);
      untagged_text[tag_ptr - string_ptr] = NUL;
      pf_output_text (untagged_text);

      /* Catch and ignore completely empty tags. */
      if (strchr (tag_ptr, GREATERTHAN) == tag_ptr + 1)
        {
          string_ptr = tag_ptr + 2;
          continue;
        }

      /* Get the tag text. */
      if (sscanf (tag_ptr, "<%[^>]>", tag) != 1)
        {
          sc_error ("pf_output_untagged: mismatched '%c'\n", LESSTHAN);

          /* Output the single < character and continue. */
          if_print_character (LESSTHAN);
          string_ptr = tag_ptr + 1;
          continue;
        }

      /* Handle tag output. */
      pf_output_tag (tag);

      /* Advance string_ptr over the <...> tag. */
      string_ptr = tag_ptr + strlen (tag) + 1;
      if (*string_ptr == GREATERTHAN)
        string_ptr++;
    }

  /* Output any remaining string text. */
  pf_output_text (string_ptr);

  /* All done. */
  sc_free (temporary_buffer);
}


/*
 * pf_filter()
 *
 * Filter output, interpolating variables and replacing ALR's, and return
 * the resulting string to the caller.  The string is malloc'ed, so the
 * caller needs to remember to free it.
 */
sc_char *
pf_filter (const sc_char *string,
           sc_var_setref_t vars, sc_prop_setref_t bundle)
{
  sc_vartype_t vt_key[1], vt_rvalue;
  sc_char *interpolated, *replaced;
  sc_int32 alr_count, index_;
  sc_bool changed, *alr_applied;
  assert (string && vars && bundle);

  /*
   * Create and clear a new set of ALR application flags.  These are used to
   * ensure that a given ALR is applied only once.
   */
  vt_key[0].string = "ALRs";
  alr_count =
    (prop_get (bundle, "I<-s", &vt_rvalue, vt_key)) ? vt_rvalue.integer : 0;
  alr_applied = sc_malloc (alr_count * sizeof (*alr_applied));

  for (index_ = 0; index_ < alr_count; index_++)
    alr_applied[index_] = FALSE;

  if (pf_trace)
    sc_trace ("Printfilter: initial \"%s\"\n", string);

  /* Interpolate until all variables are out of the string. */
  interpolated = pf_interpolate (string, vars, &changed);
  while (changed)
    {
      sc_char *target;

      if (pf_trace)
        {
          sc_trace ("Printfilter: interpolation \"%s\"\n", interpolated);
        }

      /* Repeat interpolation on last string returned. */
      target = interpolated;
      interpolated = pf_interpolate (target, vars, &changed);
      sc_free (target);
    }

  if (pf_trace)
    {
      sc_trace ("Printfilter: interpolated 1 \"%s\"\n", interpolated);
    }

  /* Replace ALRs in the current interpolated string. */
  replaced = pf_replace_alrs (interpolated, bundle, &changed, alr_applied);
  sc_free (interpolated);

  if (pf_trace)
    sc_trace ("Printfilter: replaced 1 \"%s\"\n", replaced);

  /* Now do the variable interpolation loop again. */
  interpolated = pf_interpolate (replaced, vars, &changed);
  while (changed)
    {
      sc_char *target;

      if (pf_trace)
        {
          sc_trace ("Printfilter: interpolation \"%s\"\n", interpolated);
        }

      /* Repeat interpolation on last string returned. */
      target = interpolated;
      interpolated = pf_interpolate (target, vars, &changed);
      sc_free (target);
    }

  /* No further need for the intermediate ALR replacement string. */
  sc_free (replaced);

  if (pf_trace)
    {
      sc_trace ("Printfilter: interpolated 2 \"%s\"\n", interpolated);
    }

  /*
   * Replace ALRs a second time in the current interpolated string.
   *
   * TODO How many times are we supposed to repeat this cycle?
   */
  replaced = pf_replace_alrs (interpolated, bundle, &changed, alr_applied);
  sc_free (interpolated);

  if (pf_trace)
    {
      sc_trace ("Printfilter: replaced 2 [final] \"%s\"\n", replaced);
    }

  /* Free ALR application flags, and return the final string. */
  sc_free (alr_applied);
  return replaced;
}


/*
 * pf_filter_for_info()
 *
 * Filter output, interpolating variables only (no ALR replacement), and
 * return the resulting string to the caller.  Used on informational strings
 * such as the game title and author.  The string is malloc'ed, so the
 * caller needs to remember to free it.
 */
sc_char *
pf_filter_for_info (const sc_char *string, sc_var_setref_t vars)
{
  sc_char *interpolated;
  sc_bool changed;
  assert (string && vars);

  if (pf_trace)
    sc_trace ("Printfilter: info initial \"%s\"\n", string);

  /* Interpolate until all variables are out of the string. */
  interpolated = pf_interpolate (string, vars, &changed);
  while (changed)
    {
      sc_char *target;

      if (pf_trace)
        {
          sc_trace ("Printfilter: info interpolation \"%s\"\n", interpolated);
        }

      /* Repeat interpolation on last string returned. */
      target = interpolated;
      interpolated = pf_interpolate (target, vars, &changed);
      sc_free (target);
    }

  if (pf_trace)
    {
      sc_trace ("Printfilter: info interpolated \"%s\"\n", interpolated);
    }

  /* Return the interpolated string. */
  return interpolated;
}


/*
 * pf_flush()
 *
 * Filter buffered data, interpolating variables and replacing ALR's, and
 * send the resulting string to the output channel.
 */
void
pf_flush (sc_printfilterref_t filter,
          sc_var_setref_t vars, sc_prop_setref_t bundle)
{
  sc_char *filtered;
  assert (filter && filter->magic == PRINTFILTER_MAGIC);
  assert (vars && bundle);

  /* See if there is any buffered data to flush. */
  if (filter->buffer_length > 0)
    {
      /*
       * Filter the buffered string, then print it untagged.  Remember to free
       * the filtered version.
       */
      filtered = pf_filter (filter->buffer, vars, bundle);
      pf_output_untagged (filtered);
      sc_free (filtered);

      /* Remove buffered data by resetting length to zero. */
      filter->buffer_length = 0;
    }
}


/*
 * pf_get_buffer()
 *
 * Return the raw, unfiltered, buffered text.  Primarily for use by the
 * debugger.  Returns NULL if no buffered data available.
 */
const sc_char *
pf_get_buffer (sc_printfilterref_t filter)
{
  assert (filter && filter->magic == PRINTFILTER_MAGIC);

  /*
   * Return buffer if filter length is greater than zero.  Note that this
   * assumes that the buffer is a nul-terminated string.
   */
  if (filter->buffer_length > 0)
    {
      assert (filter->buffer[filter->buffer_length] == NUL);
      return filter->buffer;
    }
  else
    return NULL;
}


/*
 * pf_buffer_string()
 * pf_buffer_character()
 *
 * Add a string, and a single character, to the printfilter buffer.
 */
void
pf_buffer_string (sc_printfilterref_t filter, const sc_char *string)
{
  sc_uint32 required, noted;
  assert (filter && filter->magic == PRINTFILTER_MAGIC);
  assert (string);

  /*
   * Calculate the required buffer size to append string.  Remember to add
   * one for the terminating NUL.
   */
  required = filter->buffer_length + strlen (string) + 1;

  /* If this is more than the current buffer size, resize it. */
  if (required > filter->buffer_size)
    {
      sc_uint32 new_size;

      /* Calculate the new malloc size, in increment chunks. */
      new_size = ((required + BUFFER_GROW_INCREMENT - 1)
                  / BUFFER_GROW_INCREMENT) * BUFFER_GROW_INCREMENT;

      /* Grow the buffer. */
      filter->buffer = sc_realloc (filter->buffer, new_size);
      filter->buffer_size = new_size;
    }

  /* If empty, put a NUL into the buffer to permit strcat. */
  if (filter->buffer_length == 0)
    filter->buffer[0] = NUL;

  /* Note append start, then append the string to the buffer. */
  noted = filter->buffer_length;
  strcat (filter->buffer, string);
  filter->buffer_length += strlen (string);

  /* Adjust first character of appended string if flagged. */
  if (filter->new_sentence)
    filter->buffer[noted] = toupper (filter->buffer[noted]);
  filter->new_sentence = FALSE;
}

void
pf_buffer_character (sc_printfilterref_t filter, sc_char character)
{
  sc_char buffer[2];
  assert (filter && filter->magic == PRINTFILTER_MAGIC);

  /* Create a short string and add this. */
  buffer[0] = character;
  buffer[1] = NUL;
  pf_buffer_string (filter, buffer);
}


/*
 * pf_new_sentence()
 *
 * Tells the printfilter to force the next non-space character to uppercase.
 */
void
pf_new_sentence (sc_printfilterref_t filter)
{
  assert (filter && filter->magic == PRINTFILTER_MAGIC);

  filter->new_sentence = TRUE;
}


/*
 * pf_buffer_tag()
 *
 * Insert an HTML-like tag into the buffered output data.
 */
void
pf_buffer_tag (sc_printfilterref_t filter, sc_uint32 tag)
{
  sc_uint16 index_;
  assert (filter && filter->magic == PRINTFILTER_MAGIC);

  /* Find the tag in the tags table. */
  for (index_ = 0; HTML_TAGS[index_].tag_string; index_++)
    {
      if (tag == HTML_TAGS[index_].tag)
        {
          /* Found a match, so output the equivalent string. */
          pf_buffer_character (filter, LESSTHAN);
          pf_buffer_string (filter, HTML_TAGS[index_].tag_string);
          pf_buffer_character (filter, GREATERTHAN);
          return;
        }
    }

  /* Tag not found in the table. */
  sc_error ("pf_buffer_tag: invalid tag, %lu\n", tag);
}


/*
 * pf_strip_tags_common()
 *
 * Strip HTML-like tags from a string.  Used to process strings used in ways
 * other than being passed to if_print_string(), for example room names and
 * status lines.  It ignores all tags except <br>, which it replaces with
 * a newline if requested by allow_newlines.
 */
static void
pf_strip_tags_common (sc_char *string, sc_bool allow_newlines)
{
  sc_char *string_ptr, *tag_start, *tag_end;

  /* Run through the string looking for <...> tags. */
  string_ptr = string;
  for (tag_start = strchr (string_ptr, LESSTHAN);
       tag_start; tag_start = strchr (string_ptr, LESSTHAN))
    {
      /* Locate tag end, and break if unterminated. */
      tag_end = strchr (tag_start, GREATERTHAN);
      if (!tag_end)
        break;

      /* If the tag is <br>, replace with newline if requested. */
      if (allow_newlines)
        {
          if (tag_end - tag_start == 3
              && sc_strncasecmp (tag_start + 1, "br", 2) == 0)
            *tag_start++ = '\n';
        }

      /* Remove the tag from the string, then advance input. */
      memmove (tag_start, tag_end + 1, strlen (tag_end));
      string_ptr = tag_start;
    }
}


/*
 * pf_strip_tags()
 * pf_strip_tags_for_hints()
 *
 * Public interfaces to pf_strip_tags_common().  The hints version will
 * allow <br> tags to map into newlines in hints strings.
 */
void
pf_strip_tags (sc_char *string)
{
  pf_strip_tags_common (string, FALSE);
}

void
pf_strip_tags_for_hints (sc_char *string)
{
  pf_strip_tags_common (string, TRUE);
}


/*
 * pf_compare_words()
 *
 * Matches multiple words from words in string.  Returns the extent of
 * the match if the string matched, 0 otherwise.
 */
static sc_uint16
pf_compare_words (const sc_char *string, const sc_char *words)
{
  sc_uint16 word_posn, posn;

  /* None expected, but skip leading space. */
  for (word_posn = 0; isspace (words[word_posn]) && words[word_posn] != NUL;)
    word_posn++;

  /* Match characters from words with the string at position. */
  posn = 0;
  while (TRUE)
    {
      /* Any character mismatch means no words match. */
      if (tolower (words[word_posn]) != tolower (string[posn]))
        return 0;

      /* Move to next character in each. */
      word_posn++;
      posn++;

      /*
       * If at space, advance over whitespace in words list.  Stop when we
       * hit the end of the words list.
       */
      while (isspace (words[word_posn]) && words[word_posn] != NUL)
        word_posn++;
      if (words[word_posn] == NUL)
        break;

      /*
       * About to match another word, so advance over whitespace in the
       * current string too.
       */
      while (isspace (string[posn]) && string[posn] != NUL)
        posn++;
    }

  /*
   * We reached the end of words.  If we're at the end of the match string,
   * or at spaces, we've matched.
   */
  if (isspace (string[posn]) || string[posn] == NUL)
    return posn;

  /* More text after the match, so it's not quite a match. */
  return 0;
}


/*
 * pf_filter_input()
 *
 * Apply synonym changes to a player input string, and return the resulting
 * string to the caller.  The string is malloc'ed, so the caller needs to
 * remember to free it.
 */
sc_char *
pf_filter_input (const sc_char *string, sc_prop_setref_t bundle)
{
  sc_vartype_t vt_key[3], vt_rvalue;
  sc_int32 synonym_count, index_;
  sc_uint16 word_index;
  sc_char *output_buffer;
  sc_uint16 output_buffer_length;
  assert (string && bundle);

  if (pf_trace)
    sc_trace ("Printfilter: input synonyms \"%s\"\n", string);

  /* Count synonyms, and set invariant part of properties key. */
  vt_key[0].string = "Synonyms";
  synonym_count =
    (prop_get (bundle, "I<-s", &vt_rvalue, vt_key)) ? vt_rvalue.integer : 0;

  /* Start off with a copy of the input string. */
  output_buffer_length = strlen (string) + 1;
  output_buffer = sc_malloc (output_buffer_length);
  strcpy (output_buffer, string);

  /* Find the first string word. */
  for (word_index = 0;
       output_buffer[word_index] != NUL && isspace (output_buffer[word_index]);)
    word_index++;

  /* Loop over each word in the string. */
  while (output_buffer[word_index] != NUL)
    {
      sc_bool matched;

      /* Loop over synonyms, looking for a match at this point. */
      matched = FALSE;
      for (index_ = 0; index_ < synonym_count; index_++)
        {
          const sc_char *original, *replacement;
          sc_uint16 extent;

          /* Set this synonym in the properties key. */
          vt_key[1].integer = index_;

          /* Retrieve the synonym original string. */
          vt_key[2].string = "Original";
          original = prop_get_string (bundle, "S<-sis", vt_key);

          /* Compare the original at this point. */
          extent = pf_compare_words (output_buffer + word_index, original);
          if (extent > 0)
            {
              sc_uint16 length, last;

              /* Matched, so find replacement. */
              vt_key[2].string = "Replacement";
              replacement = prop_get_string (bundle, "S<-sis", vt_key);
              length = strlen (replacement);

              /*
               * If necessary, grow the output buffer for the replacement.  At
               * the same time, note the last character index for the move.
               */
              if (length > extent)
                {
                  output_buffer_length += length - extent;
                  output_buffer = sc_realloc (output_buffer,
                                              output_buffer_length);
                  last = length;
                }
              else
                last = extent;

              /* Insert the replacement string. */
              memmove (output_buffer + word_index + length,
                       output_buffer + word_index + extent,
                       output_buffer_length - word_index - last);
              memcpy (output_buffer + word_index, replacement, length);

              /* Adjust the word index and break. */
              word_index += length;
              matched = TRUE;
              break;
            }
        }

      /* If no match, advance over the unmatched word. */
      if (!matched)
        {
          while (output_buffer[word_index] != NUL
                 && !isspace (output_buffer[word_index]))
            word_index++;
        }

      /* Find the next word start. */
      while (output_buffer[word_index] != NUL
             && isspace (output_buffer[word_index]))
        word_index++;
    }
  while (output_buffer[word_index] != NUL);

  if (pf_trace)
    sc_trace ("Printfilter: input filtered \"%s\"\n", output_buffer);

  /* Return the final string. */
  return output_buffer;
}


/*
 * pf_debug_trace()
 *
 * Set filter tracing on/off.
 */
void
pf_debug_trace (sc_bool flag)
{
  pf_trace = flag;
}
